/*
 *     Copyright CERN, Geneva 1989 - Copyright and any other
 *     appropriate legal protection of these computer programs
 *     and associated documentation reserved in all countries
 *     of the world.
 */

#include "telnet.h"
#include "3270.h"
#include "globals.h"

/* variables to note new values found */
static int field; /* first position of newly found field */
static int color; /* color associated with the newfield */

static          int      yval, xval;     /* position in image line */
static          short    amwrap = FALSE; /* set TRUE in autowrap last col write */

static          int      copy_n;
static          int      mvflag = FALSE;

static          unsigned short          scrchar;

unsigned char   *chsp;          /* pointer to curscr current line character */
unsigned char   *ccssp;         /* pointer to curscr current line color and character set */
int             screeni;        /* index into screen */
int             nlsp;           /* Number of chars up to last non-space char */

#ifdef LDEBUG
	char    *colorname[MAXCOLORS] =  { " NORMAL ",
					   " NORM_BLINK ",
					   " NORM_PROT ",
					   " NORM_PR_BLINK ",
					   " BOLD ",
					   " BOLD_BLINK ",
					   " BOLD_PROT ",
					   " BOLD_PR_BLINK ",
					   " NORM_MOD ",
					   " NORM_M_BLINK ",
					   " BOLD_MOD ",
					   " BOLD_M_BLINK ",
					   " STATL ",
					   " STATL_BLINK ",
					   " INVISIBLE "
					 };
#endif /* LDEBUG */

int     colortab[128] = { NORMAL, NORMAL, NORM_MOD, NORM_MOD,
			  NORM_PROT, NORM_PROT, NORM_PROT, NORM_PROT,
			  BOLD, BOLD, BOLD_MOD, BOLD_MOD,
			  INVISIBLE, INVISIBLE, INVISIBLE, INVISIBLE,
			  NORMAL, NORMAL, NORM_MOD, NORM_MOD,
			  NORM_PROT, NORM_PROT, NORM_PROT, NORM_PROT,
			  BOLD, BOLD, BOLD_MOD, BOLD_MOD,
			  INVISIBLE, INVISIBLE, INVISIBLE, INVISIBLE,
			  NORM_PROT, NORM_PROT, NORM_PROT, NORM_PROT,
			  NORM_PROT, NORM_PROT, NORM_PROT, NORM_PROT,
			  BOLD_PROT, BOLD_PROT, BOLD_PROT, BOLD_PROT,
			  INVISIBLE, INVISIBLE, INVISIBLE, INVISIBLE,
			  NORM_PROT, NORM_PROT, NORM_PROT, NORM_PROT,
			  NORM_PROT, NORM_PROT, NORM_PROT, NORM_PROT,
			  BOLD_PROT, BOLD_PROT, BOLD_PROT, BOLD_PROT,
			  INVISIBLE, INVISIBLE, INVISIBLE, INVISIBLE,
			  NORM_BLINK, NORM_M_BLINK, NORM_BLINK, NORM_M_BLINK,
			  NORM_BLINK, NORM_BLINK, NORM_BLINK, NORM_BLINK,
			  BOLD_BLINK, BOLD_M_BLINK, BOLD_BLINK, BOLD_M_BLINK,
			  INVISIBLE, INVISIBLE, INVISIBLE, INVISIBLE,
			  NORM_BLINK, NORM_M_BLINK, NORM_BLINK, NORM_M_BLINK,
			  NORM_BLINK, NORM_BLINK, NORM_BLINK, NORM_BLINK,
			  BOLD_BLINK, BOLD_M_BLINK, BOLD_BLINK, BOLD_M_BLINK,
			  INVISIBLE, INVISIBLE, INVISIBLE, INVISIBLE,
			  NORM_PR_BLINK, NORM_PR_BLINK, NORM_PR_BLINK, NORM_PR_BLINK,
			  NORM_PR_BLINK, NORM_PR_BLINK, NORM_PR_BLINK, NORM_PR_BLINK,
			  BOLD_BLINK, BOLD_BLINK, BOLD_BLINK, BOLD_BLINK,
			  INVISIBLE, INVISIBLE, INVISIBLE, INVISIBLE,
			  NORM_PR_BLINK, NORM_PR_BLINK, NORM_PR_BLINK, NORM_PR_BLINK,
			  NORM_PR_BLINK, NORM_PR_BLINK, NORM_PR_BLINK, NORM_PR_BLINK,
			  BOLD_BLINK, BOLD_BLINK, BOLD_BLINK, BOLD_BLINK,
			  INVISIBLE, INVISIBLE, INVISIBLE, INVISIBLE
			};


/*
 * Set up all of the strings to change from one "color" to another.
 * Set up the pointer array change_cp[from][to]
 * and the string lengths in change_sl[from][to]
 */

int     ifrom, ito;
int     chbufsiz;
bool    count_only;

#ifdef STANDARD_C

int     add_char(char addch)

#else /* STANDARD_C */

int     add_char(addch)

char    addch;

#endif /* STANDARD_C */

{
	if (!count_only)
	{
		change_cp[ifrom][ito][change_sl[ifrom][ito]] = (addch ? addch : 0x80);
	}
	chbufsiz++;
	change_sl[ifrom][ito] += 1;
	return(0);
}

#ifdef STANDARD_C

void    add_string(int from, int to, char *tcstring)

#else /* STANDARD_C */

void    add_string(from, to, tcstring)

int     from, to;
char    *tcstring;

#endif /* STANDARD_C */

{
	ifrom = from;
	ito = to;
	tputs(tcstring , 1 , add_char);
}

#ifdef STANDARD_C

void    copy_cp(int from, int to, int oldfrom, int oldto)

#else /* STANDARD_C */

void    copy_cp(from, to, oldfrom, oldto)

int     from, to;
int     oldfrom, oldto;

#endif /* STANDARD_C */

{

	if (!count_only)
	{
		(void)strcpy(change_cp[from][to] + change_sl[from][to],
			     change_cp[oldfrom][oldto]);
	}
	change_sl[from][to] += change_sl[oldfrom][oldto];
	chbufsiz += change_sl[oldfrom][oldto];
}

#ifdef STANDARD_C

void    cchange_setup(void)

#else /* STANDARD_C */

void    cchange_setup()

#endif /* STANDARD_C */

{
	int     from, to, i;
	char    *chbufad = NULL;
	char    *chbufpt = NULL;
	int     pass_number;

#ifdef LDEBUG
	char    *ch;
#endif /* LDEBUG */

	chbufsiz = 0;
	for (pass_number = 0 ; pass_number < 4 ; pass_number++)
	{
		switch (pass_number)
		{
			case 0:
				count_only = TRUE;
				break;

			default:

#ifdef LDEBUG
	LOG(log, "Before pass %d \n", pass_number);
	for (from = 0 ; from < MAXCOLORS ; from++)
		for (to = 0 ; to < MAXCOLORS ; to++)
		{
			if (change_sl[from][to])
			{
				LOG(log, "from %s to %s string (%d chars) is <",
					colorname[from], colorname[to],
					change_sl[from][to]);

				if (ch = change_cp[from][to])
					for (i = 0 ; i < change_sl[from][to] ; i++)
					{
						if (!(*ch))
							break;
						LOG(log, "%s", print_ascii((unsigned short)*ch));
						ch++;
					}
				LOG(log, ">\n");
			}
		}
#endif /* LDEBUG */

				free((void *) chbufad);

			case 1:
				chbufpt = chbufad = (char *) alloc_buffer(chbufsiz, '\0');
				chbufsiz = 0;
				count_only = FALSE;
				break;
		}

		for (from = 0 ; from < MAXCOLORS ; from++)
		{
			if (count_only)
			{
				move_ok[from] = TRUE;
				ceol_ok[from] = TRUE;
			}
			for (to = 0 ; to < MAXCOLORS ; to++)
			{
				if (!count_only)
				{
					change_cp[from][to] = chbufpt;
					chbufpt += (change_sl[from][to] + 1);
				}
				chbufsiz++;
				change_sl[from][to] = 0;
			}
		}

		/* temporarily need to know the ME sequence */
		add_string(NORMAL , NORMAL , ME);

		/* Try to use bright mode */

		if (MD)
		{
			add_string(NORMAL , BOLD , MD) ;
			add_string(BOLD , NORMAL , ME);
			add_string(NORMAL , BOLD_PROT , MD);
			add_string(BOLD_PROT , NORMAL , ME);

			/*
			 * On a real VT340 a bright blank is different from a normal blank
			 * and so we cannot do any "clear to end of line" for bright
			 * blanks. However, this slows things down a lot, especially
			 * in something like FILELIST on a 132-column line.
			 * I have therefore commented it out!
			 */

		     /* ceol_ok[BOLD] = FALSE;      */
		     /* ceol_ok[BOLD_PROT] = FALSE; */
		}

		/* Try to use reverse video mode */

		if (MR)
		{
			add_string(NORMAL , BOLD_MOD , MR);
			add_string(BOLD_MOD , NORMAL , ME);
			ceol_ok[BOLD_MOD] = FALSE;
		}

		/* Try to use underscore mode */

		if (US)
		{
/*                      move_ok[NORM_MOD] = move_ok[NORM_M_BLINK] = MS; */
/*                      add_string(NORMAL , NORM_MOD , US);       */
/*                      add_string(NORM_MOD , NORMAL , UE);       */
		}

		/* Maybe the termcap defines standout mode for status line */

		if (SO)
		{
			move_ok[STATL] = move_ok[STATL_BLINK] = MS;
			add_string(NORMAL , STATL , SO);
			add_string(STATL , NORMAL , SE);
			if (MR && !strcmp(MR, SO))
				ceol_ok[STATL] = FALSE;
		}

		/* That is enough for the first pass */

		if (pass_number < 1)
			continue;

		/* On the second pass we need to switch to counting only */

		if (pass_number == 1)
			count_only = TRUE;

		/* Try to use blinking mode */

		if (MB)
		{
			ceol_ok[NORM_BLINK] = FALSE;
			add_string(NORMAL , NORM_BLINK , MB);
			add_string(NORM_BLINK , NORMAL , ME);
		}

		for (i = NORMAL+2 ; i < INVISIBLE ; i += 2)
		{
			copy_cp(NORMAL , i+1 , NORMAL, i);

			if (MB)
			{
				ceol_ok[i+1] = FALSE;
				add_string(NORMAL , i+1 , MB);
				add_string(i+1 , NORMAL , ME);
			}
			else
				copy_cp(i+1 , NORMAL , i , NORMAL);

			copy_cp(NORMAL+1 , i , NORMAL+1, NORMAL);
			copy_cp(NORMAL+1 , i , NORMAL, i);

			if (MB)
			{
				if (strcmp(change_cp[NORMAL+1][NORMAL] , change_cp[NORMAL][NORMAL]))
				{
					copy_cp(NORMAL+1 , i+1 , NORMAL+1 , NORMAL);
					copy_cp(NORMAL+1 , i+1 , NORMAL , i+1);
					add_string(NORMAL+1 , i+1 , MB);
				}
				else
					copy_cp(NORMAL+1 , i+1 , NORMAL, i);
			}
			else
				copy_cp(NORMAL+1 , i+1 , NORMAL, i);

			copy_cp(i , NORMAL+1 , i, NORMAL);
			if (MB)
			{
				add_string(i , NORMAL+1 , MB);

				if (strcmp(change_cp[i][NORMAL] , change_cp[NORMAL][NORMAL]))
				{
					copy_cp(i+1 , NORMAL+1 , i, NORMAL);
				}
				else
				{
					add_string(i+1 , NORMAL+1 , ME);
					add_string(i+1 , NORMAL+1 , MB);
				}
			}
			else
				copy_cp(i+1 , NORMAL+1 , i, NORMAL + 1);

		}
		if (MB)
		{
			add_string(INVISIBLE , NORM_BLINK , MB);
			add_string(NORM_BLINK , INVISIBLE , ME);
		}

		/* That is enough for the second pass */

		if (pass_number < 2)
			continue;

		/* On the third pass we need to switch to counting only */

		if (pass_number == 2)
			count_only = TRUE;

		for (from = NORMAL+2 ; from <= INVISIBLE ; from += 1)
		{
			for (to = NORMAL+2 ; to <= INVISIBLE ; to += 1)
			{
				if (strcmp(change_cp[NORMAL][from] , change_cp[NORMAL][to]))
				{
					if ((from & 0x1) || (to != from+1))
					{
						copy_cp(from , to , from, NORMAL);
						copy_cp(from , to , NORMAL, to);
					}
					else
					{
						if (MB)
							add_string(from , to , MB);
					}
				}
			}
		}
	}

	/* reset NORMAL to NORMAL */
	for (i = 0 ; i < change_sl[NORMAL][NORMAL] ; i++)
		change_cp[NORMAL][NORMAL][i] = '\0';
	change_sl[NORMAL][NORMAL] = 0;

#ifdef LDEBUG
	LOG(log, "After %d passes\n", pass_number);
	for (from = 0 ; from < MAXCOLORS ; from++)
		for (to = 0 ; to < MAXCOLORS ; to++)
		{
			if (change_sl[from][to])
			{
				LOG(log, "from %s to %s string (%d chars) is <",
					colorname[from], colorname[to],
					change_sl[from][to]);

				if (ch = change_cp[from][to])
					for (i = 0 ; i < change_sl[from][to] ; i++)
					{
						if (!(*ch))
							break;
						LOG(log, "%s", print_ascii((unsigned short)*ch));
						ch++;
					}
				LOG(log, ">\n");
			}
		}
#endif /* LDEBUG */


}

/*
 * Calculate how many output chars would be needed to move from current
 * position, given by ly and lx, to new position.
 * the move might be done by either cursor movement (CM) or the special
 * VT-type cursor right sequence (which is used in fgoto inside mvcur).
 * Remember that after an autowrap at the end of the line we should use
 * straight cursor movement.
 */

#ifdef STANDARD_C

extern  int     movestrcnt(int ny, int nx)

#else /* STANDARD_C */

int     movestrcnt(ny, nx)
int     ny, nx;

#endif /* STANDARD_C */

{
	return (strlen(tgoto(CM, nx, ny)));
}

/*
 * perform a mvcur, returning to NORMAL mode if necessary
 */

#ifdef STANDARD_C

void    domvcur(int oy, int ox, int ny, int nx)

#else /* STANDARD_C */

void    domvcur(oy, ox, ny, nx)

int     oy, ox, ny, nx;

#endif /* STANDARD_C */

{

#ifdef LDEBUG
	LOG(log, "domvcur         : (%d %d to %d %d) amwrap is %d\n",
		  oy, ox, ny, nx, amwrap);
#endif /* LDEBUG */

	if (!move_ok[curscr->_color])
	{

#ifdef LDEBUG
	LOG(log, "domvcur         : Go to NORMAL mode, flags = %s\n",
		  colorname[curscr->_color]);
#endif /* LDEBUG */

		f_puts(change_cp[curscr->_color][NORMAL], FALSE);
		curscr->_color = NORMAL;
	}

	/* after writing last col on autowrap force a real move */
	if (amwrap)
	{
		amwrap = FALSE;
	}
	else
	{

		/*
		 * In some cases we can find "efficient" cursor move sequences.
		 * However, this can be dangerous for the case where the
		 * termcap sequence for "do" is line feed: some systems seem to
		 * add a carriage return after it when in CBREAK mode.
		 * For workstations (TTY not defined), therefore, we should not look for
		 * particularly short sequences
		 */

#ifdef TTY

		if (ny == oy)
		{
			/* move on same line */

			if (nx > ox)
			{
				if ((nx == (ox+1)) && tcslen[tcsND])
				{
					f_puttcs(tcsND, FALSE);
					return;
				}
				if (RT)
				{
					tputs(tgoto(RT, nx-ox, nx-ox), 1, f_putcs);
					return;
				}
			}
			if ((nx == (ox-1)) && tcslen[tcsLE])
			{
				f_puttcs(tcsLE, FALSE);
				return;
			}
		}
		if ((nx == ox)  && !scroll_line)
		{
			/* move on same column */
			if ((ny == (oy+1)) && tcslen[tcsDO])
			{
				f_puttcs(tcsDO, FALSE);
				return;
			}
			if ((ny == (oy-1)) && tcslen[tcsUP])
			{
				f_puttcs(tcsUP, FALSE);
				return;
			}
		}
		if ((nx == 0) && tcslen[tcsCR] && !scroll_line)
		{
			if ((ny == (oy+1)) && tcslen[tcsDO])
			{
				f_puttcs(tcsDO, FALSE);
				f_puttcs(tcsCR, FALSE);
				return;
			}
			if (ny == oy)
			{
				f_puttcs(tcsCR, FALSE);
				return;
			}
		}

#endif /* TTY */

	}
	tputs(tgoto(CM, nx, ny), 1, f_putcs);
}

/*
 * Check if two characters will be seen as different on the user screen.
 *
 * On entry we expect chsp to point to the character already on the user
 * screen and ccssp to point to its color and character set.
 * screeni will be an index of the (16-bit) character in the screen array
 * which we wish to have on the user screen.
 * A TRUE return says that the characters are different, whilst a FALSE
 * return says they are the same (in the sense that they will look the same
 * on the user's screen).
 * A by-product of the checking is to set scrchar to be the 16-bit short
 * consisting of (from the most significant end):-
 *
 * 4 bits for the color
 * 4 bits for the character set and Graphics Escape indicators
 * 8 bits for the ASCII 8-bit character
 */

#ifdef STANDARD_C

bool    isdiff(void)

#else /* STANDARD_C */

bool    isdiff()

#endif /* STANDARD_C */

{
	register short          c;
	register short          ch2;
	register int            chcolor;

	/*
	 * Given an index into screen, find what character is to go into
	 * the curses structure and onto the terminal.
	 * The most simple case is a terminal that is black and white, such that
	 * we are only interested in the highBIT (for standout mode).
	 * If the terminal can do blinking then we might also use the blinkingBIT.
	 * A colour screen might also be interested in the protectedBIT.
	 * Note that on entry color is set, but it might be reset here.
	 */

	chcolor = color;
	if (c = screen[screeni])
	{
		/* check for a field definition character */

		if (c & FIELDSTART)
		{
			field = screeni;
			color = colorbits((int)c);
			chcolor = NORMAL;

#ifdef LDEBUG
	LOG(log, "isdiff          : new field at (%d,%d), color %s\n",
		       YX(screeni), colorname[color]);
#endif /* LDEBUG */

			ch2 = ' ';
		}
		/* otherwise, tint c appropriately */
		/* all other chars, including blanks, come with color */

		else
			ch2 = (chcolor == INVISIBLE ? (short)' ' : (c & 0xfff));
	}
	else
		ch2 = ' ';

	scrchar = (chcolor << 12) | ch2;

	if (*chsp != (unsigned char)ch2)
		return (TRUE);
	if (((*ccssp) & 0xf) != (ch2 >> 8))
		return (TRUE);
	if (!change_sl[(*ccssp) >> 4][chcolor])
	{
		*ccssp = ((*ccssp) & 0x0f) + (chcolor << 4);
		return (FALSE);
	}
	if ((ch2 == (short)' ')      &&
	    (ceol_ok[(*ccssp) >> 4]) &&
	    (ceol_ok[chcolor]))
		return (FALSE);
	return (TRUE);
}

/*
 * makeline(scr1,scr2) is called with scr1 and scr2 as pointers to a full or
 * partial line in screen: they are the first and last pointers.
 * It is assumed that the relevant field start to screen[scr1] is field
 * and the corresponding color is color: of course, scr1 can be the
 * field start itself.
 * However, the actual cursor position, current field start and current
 * color are in cursorpos, fieldstart and currentcolor.
 * If output is actually done (because something has changed) then
 * we update these variables.
 * Be careful to avoid filling in the very last screen character if on a
 * real terminal where the screen size matches exactly the 3270
 * screen size and wrap mode is on.
 * In fact, on some emulators this can be done because the cursor appears
 * to stay on the last character of a line, rather than causing a warp.
 * However, this raises extra complications if we then try a cursor movement
 * such as up, or a clear to end of line!
 */

#ifdef STANDARD_C

void    makeline(int scr1, int scr2)

#else /* STANDARD_C */

void    makeline(scr1, scr2)

int     scr1,scr2;

#endif /* STANDARD_C */

{
	unsigned char           *chsp1, *ccssp1;
	unsigned short          prtchar;


	register int            lch;
	int                     lchlast;
	int                     i1;
	int                     xval1;
	int                     orig_color, cur_color;
	int                     puts_n;
	int                     savefield, savecolor;
	int                     prevfield, prevcolor;
	int                     charset;

	yval = scr1/linelen;
	xval = scr1%linelen;
	chsp = char_image[yval] + xval;
	ccssp = ccset_image[yval] + xval;
	lch = scr2%linelen;

#ifdef LDEBUG
	LOG(log, "makeline        : entry (%d,%d) to (%d,%d)\n",
	       YX(scr1),YX(scr2));
	LOG(log, "                : field (%d,%d) color %s\n",
	       YX(field),colorname[color]);
#endif /* LDEBUG */

	/* skip to the first screen character which has changed */
	for (screeni = scr1 ; screeni <= scr2 ; screeni++)
	{
		if (((screen[screeni] != (short)*chsp) || (color != ((*ccssp) >> 4))) &&
		    (isdiff()))
			break;
		chsp++;
		ccssp++;
	}
	if (screeni > scr2)
	{
		/* nothing to put onto the user screen */
		return;
	}

	/* note the current value of field and color */
	/* because isdiff() can reset field and color */

	savefield = field;
	savecolor = color;

	xval1 = (i1 = screeni)%linelen;
	chsp1 = chsp;
	ccssp1 = ccssp;

#ifdef LDEBUG
	LOG(log, "makeline        : changed (%d,%d): xval1 = %d\n",
		   YX(i1), xval1);
	LOG(log, "makeline        : '%s' %s",
		   print_ascii(((unsigned short)((*ccssp) & 0x0f) << 8) | (short)*chsp), colorname[(*ccssp) >> 4]);
	LOG(log, " to '%s' %s\n",
		   print_ascii(scrchar), colorname[scrchar >> 12]);
#endif /* LDEBUG */

	/*
	 * Now, we normally will have to move the cursor to the first
	 * changed character of this line, i.e. xval1.
	 * However, in cases where we have previously written the last
	 * character of the previous line (which has COLS columns, remember)
	 * and the terminal has autowrap set
	 * and we are updating from the start of the new line (i.e. scr1
	 * is the start of this line, so that xval is zero)
	 * then maybe we don't need to do this.
	 * Remember, though, that we cannot use an escape sequence
	 * without first moving the cursor.
	 * Of course, if xval1 is > zero we will need to copy the first
	 * few characters again, which is only worthwhile if the number
	 * of characters is not larger than a cursor move sequence.
	 * The length of such a move is got from movestrcnt.
	 */

	if ((amwrap) &&
	    (ly == yval) &&
	    (!xval) &&
	    (movestrcnt(yval, xval1) > xval1))
	{
		/* force output from start of line */
		copy_n = xval1;

#ifdef LDEBUG
	LOG(log, "makeline        : autowrap line: yval =%d, copy_n = %d\n",
		       yval, copy_n);
#endif /* LDEBUG */

		/* and force out at least one character */
		if (!copy_n) copy_n = 1;

#ifdef LDEBUG
	LOG(log, "makeline        : autowrap line: force copy_n = %d\n",
		       copy_n);
#endif /* LDEBUG */

		xval1 = 0;
		chsp = chsp1 = char_image[yval];
		ccssp = ccssp1 = ccset_image[yval];
		i1 = scr1;
		mvflag = FALSE;
	}
	else
	{
		copy_n = 0;
		mvflag = TRUE;
	}

	/*
	 * Now do a first scan of the remaining characters to see
	 * where is the last character which has changed, updating
	 * lch to be the index of this character in screen.
	 * Also set nlsp to be the index of the last character
	 * in the line after which we can use the CE sequence for clear
	 * to end of line: if CE is null then nlsp is so large as never
	 * to be exceeded (i.e. the line length).
	 * In other words, output the first lch+1 characters.
	 * Of course, this number of characters could end before,
	 * in or after the region which is (possibly) modified.
	 * Also, must be careful if the line includes field
	 * definitions (except for a field definition occurring
	 * as the last character of the line.
	 * This is actually an interesting case, because on exit the
	 * values of field and color have to be taken from it.
	 */

	xval = xval1;
	lchlast = lch;
	nlsp = xval1 - 1;

	for (screeni = i1 ; screeni <= scr2 ; screeni++)
	{
		if ((((scrchar = screen[screeni]) != (short)*chsp) || (color != ((*ccssp) >> 4))) &&
		    (isdiff()))
			lchlast = xval;

		if (((scrchar & 0xff) != (unsigned short) ' ') ||
		    (!ceol_ok[scrchar >> 8]))
			nlsp = xval;

		chsp++;
		ccssp++;
		xval++;
	}
	lch = lchlast;

	/* maybe do not put out last character */
	if ((lch == (COLS - 1)) && (yval == (LINES - 1)) && AM & !VW)
		lch--;

	if (CE)
	{
		/* continue scanning of what user is seeing */
		for ( ; xval < COLS ; xval++)
		{
			if ((*chsp != ' ') ||
			    (!ceol_ok[(*ccssp) >> 4]))
				nlsp = xval;
			chsp++;
			ccssp++;
			xval++;
		}
	}
	else
		nlsp = linelen;

	/* The scanning with isdiff() may have set the field and color
	 * variables to the last field definition in the line.
	 * if so then remember them, because they should be set on exit.
	 */
	prevfield = field;
	prevcolor = color;

	/* now reset field and color where we were */
	field = savefield;
	color = savecolor;
	screeni = i1;
	xval = xval1;
	chsp = chsp1;
	ccssp = ccssp1;

#ifdef LDEBUG
	LOG(log, "makeline        : xval1 = %d, lch = %d, nlsp = %d\n",
		       xval1, lch, nlsp);
#endif /* LDEBUG */

	while (xval <= lch)
	{
		/* Each time we come here we have several variables set:-
		 *
		 * copy_n       No. of chars to copy even if unchanged
		 * mvflag       TRUE if we need to move the cursor
		 */

		if ((isdiff()) ||
		    (copy_n)                 ||
		    (xval > nlsp))
		{
			if (mvflag)
			{
				domvcur(ly, lx, yval, xval);
				mvflag = FALSE;
			}

#ifdef LDEBUG
	LOG(log, "makeline        : start copy: xval = %d, copy_n = %d\n",
		       xval, copy_n);
#endif /* LDEBUG */

			ly = yval;
			lx = xval;
			while (xval <= lch)
			{
			/* copy across at least copy_n characters */
			/* otherwise stop when the char is already on the screen */

				if (xval > nlsp)
				{
					if (((lch - xval) > tcslen[tcsCE]) &&
					    (!copy_n))
					{

#ifdef LDEBUG
	LOG(log, "makeline        : using CE: xval = %d\n",xval );
#endif /* LDEBUG */

						f_puttcs(tcsCE, FALSE);
						lx = xval;
						(void) memset((void *)chsp, ' ' , (size_t)(COLS-xval));
					     /* chsp += (COLS-xval); */
						(void) memset((void *) ccssp, 0, (size_t)(COLS-xval));
					     /* ccssp += (COLS-xval); */
					}
					else
					{

#ifdef LDEBUG
	LOG(log, "makeline        : No need to use CE: xval = %d\n",xval );
#endif /* LDEBUG */

						if (copy_n <= (lch - xval))
							copy_n = lch - xval + 1;
					}
				}

				if ((!isdiff()) &&
				    (!copy_n))
				{

#ifdef LDEBUG
	LOG(log, "makeline        : same char '%s' %s at xval = %d, lx = %d\n",
			print_ascii((unsigned short)(((*ccssp) & 0x0f) << 8) | (short)*chsp), colorname[(*ccssp) >> 4], xval, lx);
#endif /* LDEBUG */

					break;
				}

				/*
				 * change screen mode to what the scrchar wants
				 */

				if ((scrchar >> 12) != curscr->_color)
				{

#ifdef LDEBUG
	LOG(log, "makeline        : change screen from %s to %s\n",
			colorname[curscr->_color],
			colorname[scrchar >> 12]);
#endif /* LDEBUG */

					f_puts(change_cp[curscr->_color][scrchar >> 12], FALSE);
					curscr->_color = scrchar >> 12;
				}

				xval++;

				/*
				 * Now we want to work out what character to print.
				 * This allows us to cater for the special cases where
				 * scrchar is not exactly printable.
				 */

				if (!((prtchar = scrchar) & GE_CHAR))
					prtchar = atop[(int) (prtchar & 0xff)];

				if ((prtchar & GE_UNDERLINE) && US)
					f_puts(US, FALSE);

				if (charset = (prtchar & GE_SET_MASK))
				{

#ifdef LDEBUG
	LOG(log, "makeline        : at xval = %d put special character 0x%x\n",
			xval, prtchar);
#endif /* LDEBUG */

					if ((charset > GE_A_8) || !eight_bc)
					{
						int     charsind;
						char    char8;

						if (charset & GE_NBG)
						{
							char8 = ((charset & GE_G_7) ? NBG_notgot : NBG_illegal);
							charsind = 0;
						}
						else
						{
							char8 = (char)prtchar;
							charsind = (charset >> GE_SET_SHIFT) & 0x03;
						}
						f_puts(SO_string[charsind], FALSE);
						f_putc(char8);
						f_puts(SI_string[charsind], FALSE);
					}
					else
						f_putc( (char) prtchar);
				}
				else
					f_putc( (char) prtchar);

				if ((prtchar & GE_UNDERLINE) && US)
					f_puts(UE, FALSE);

				*chsp++ = scrchar;
				*ccssp++ = (scrchar >> 8);
				amwrap = FALSE;
				screeni++;
				if (copy_n) --copy_n;
			}

#ifdef LDEBUG
	LOG(log, "makeline        : end of copying at xval = %d, lx = %d\n",
			xval, lx);
#endif /* LDEBUG */

			if (lx == xval) /* if no change */
				break;
			lx = xval;

			/*
			 * must deal with case where we are at the end of
			 * the line, and maybe we have not wrapped around.
			 * In fact, there is a frig if using a terminal
			 * emulator which "stays" on the last column.
			 */
			if (lx == COLS)
			{
				if (AM)
				{
					lx = 0;
					ly++;
					amwrap = TRUE;

#ifdef LDEBUG
	LOG(log, "makeline        : autowrap to ly = %d, lx = %d\n",
			ly , lx);
#endif /* LDEBUG */

				}
				else
				{
					lx--;

#ifdef LDEBUG
	LOG(log, "makeline        : no autowrap: reset  lx = %d\n", lx);
#endif /* LDEBUG */

				}
			}
		}
/*
 * This "else" part rewritten by JMG.
 * Include code to count how many bytes would be needed to output the characters
 * up to and including the next different character.
 * If this is less than the number of bytes which would be needed by a cursor
 * move then backspace xval over these characters and return to output them.
 * Need to take into account moves into and out of standout mode, as well
 * as underscores.
 * Use of SO/SI also counts if eight-bit character output is not allowed.
 */

		else if (xval < lch)
		{

#ifdef LDEBUG
	LOG(log, "makeline        ; could skip from xval = %d, lx = %d\n",
		       xval, lx);
#endif /* LDEBUG */

			orig_color = cur_color = curscr->_color;
			puts_n = 0;

#ifdef LDEBUG
	LOG(log, "makeline        : set orig_color = 0x%x, puts_n = %d\n",
				    orig_color, puts_n);
#endif /* LDEBUG */

			/* save values in case we decide to backtrack */
			savefield = field;
			savecolor = color;

			while (xval <= lch)
			{
				if (!mvflag)
				{
					unsigned short          prntchar;

					puts_n += change_sl[cur_color][color];
					cur_color = color;

					if (!((prntchar = scrchar) & GE_CHAR))
						prntchar = atop[(int) (prntchar & 0xff)];

					if ((prntchar & GE_UNDERLINE) && US)
						puts_n += (strlen(US) + strlen(UE));

					if (charset = (prntchar & GE_SET_MASK))
					{
						if ((charset > GE_A_8) || !eight_bc)
						{
							int     charsind;

							if (charset & GE_NBG)
								charsind = 0;
							else
								charsind = (charset >> GE_SET_SHIFT)& 0x03;

							puts_n += (strlen(SO_string[charsind]) + strlen(SI_string[charsind]));
						}
					}

					++puts_n;

					if ((isdiff()) ||
					    (xval > nlsp))
					{

#ifdef LDEBUG
	LOG(log, "makeline        : now change character at xval = %d, puts_n = %d\n",
			  xval, puts_n);
	LOG(log, "makeline        : '%s' %s",
		   print_ascii(((unsigned short)((*ccssp) & 0x0f) << 8) | (short)*chsp), colorname[(*ccssp) >> 8]);
	LOG(log, " to '%s' %s\n",
		   print_ascii(scrchar), colorname[scrchar >> 12]);
#endif /* LDEBUG */

						/* now check against length of cursor move string */
						/* allowing for any necessary screen style change */

						if (!move_ok[orig_color])
						{
							puts_n -= change_sl[orig_color][NORMAL];
							orig_color = NORMAL;
						}
						puts_n -= change_sl[orig_color][cur_color];

						if (puts_n < movestrcnt(yval, xval))
						{
							/* better to copy than move cursor */
							xval -= copy_n;
							screeni -= copy_n;
							chsp -= copy_n;
							ccssp -= copy_n;
							field = savefield;
							color = savecolor;

#ifdef LDEBUG
	LOG(log, "makeline        : go back to xval = %d, copy_n = %d\n",
		       xval, copy_n);
#endif /* LDEBUG */

						}
						else
						{
							copy_n = 0;
							mvflag = TRUE;
						}
						break;
					}
					++copy_n;

				}
				else
				{
					if (isdiff())
						break;
				}
				screeni++;
				chsp++;
				ccssp++;
				++xval;
			}
		}
		else
			break;
	}
	field = prevfield;
	color = prevcolor;
	return;
}

/*
 * clear the curscr, possibly even going to the alternate screen
 */

#ifdef STANDARD_C

void    clscreen(void)

#else /* STANDARD_C */

void    clscreen()

#endif /* STANDARD_C */

{
	if (clearsn > 0)
	{
		/* have to clear screen first */
		/* may even have to swap to alternate screen */

		/* check if we currently have the wrong screen */
		if (altscreen != clearsn - 1)
		{

#ifdef DEBUG
	(void)fprintf(outf,"clscreen    : Flip to screen %d : COLS = %d, linelen = %d\n",
			clearsn-1 , COLS , linelen);
#endif /* DEBUG */

			/* flip to synchronous mode for important output */

			reset_tty(-SYNCH_NORMAL, TRUE);

			if (change_sl[curscr->_color][NORMAL])
			{
				f_puts(change_cp[curscr->_color][NORMAL], FALSE);
				curscr->_color = NORMAL;
			}
			color = NORMAL;
			slsave();
			delwin();
			if (altscreen)
			{
				LINES = normal_LINES;
				COLS = normal_COLS;
				linelen = normal_linelen;
				rows = normal_rows;
			}
			else
			{
				LINES = altern_LINES;
				COLS = altern_COLS;
				linelen = altern_linelen;
				rows = altern_rows;
			}
			while ((curscr = newwin()) == ERR)
				(void)sleep(5);
			char_image = curscr->char_y;
			ccset_image = curscr->ccset_y;

			free((void *)screen);
			free((void *)s_point);
			screensize = linelen * rows;
			screen = (short *) alloc_buffer((int)(screensize * sizeof(short)), '\0');
			s_point = (short *) alloc_buffer((int)(screensize * sizeof(short)), '\0');
			if (altflip)
			{
				f_puttcs((altscreen ? tcsSN : tcsSW), TRUE);
			}
			clearnormal();

#ifdef DEBUG
	(void)fprintf(outf,"clscreen    : Flip to screen %d\n", clearsn-1);
#endif /* DEBUG */

/*
			reset_tty(SYNCH_NORMAL, TRUE);
			sizer();
			if ((COLS != (altscreen ? normal_COLS : altern_COLS)) ||
			    (LINES != (altscreen ? normal_LINES : altern_LINES)))
				end_program(222, "Somehow I cannot switch back to the other screen");
			}
 */

			/* Flip back to normal asynch mode now */

			setup_tty(FALSE);

			if (altscreen)
			{
				COLS = normal_COLS;
				LINES = normal_LINES;
				linelen = normal_linelen;
				rows = normal_rows;
				slinenum = normal_slinenum;
			}
			else
			{
				COLS = altern_COLS;
				LINES = altern_LINES;
				linelen = altern_linelen;
				rows = altern_rows;
				slinenum = altern_slinenum;
			}

			sline = (slinenum ? TRUE : HS);
			slrestore();

#ifdef DEBUG
	(void)fprintf(outf,"clscreen    : Flipped to screen %d : COLS = %d, linelen = %d\n",
			clearsn-1 , COLS , linelen);
#endif /* DEBUG */

			altscreen = clearsn - 1;
		}
		clearscreen();
		cursorpos = bufferpos = 0;
		clearsn = -clearsn;
		normal_chars = normal_blanks = TRUE;
	}
}

/*
 * update curscr from screen[startpos] to screen[endpos]
 * note that since this can be called several times in a row, it does
 * not set the current cursor: this must be done in a subsequent call
 * of putcursor() or writecursor().
 */

#ifdef STANDARD_C

void    update_curscr(void)

#else /* STANDARD_C */

void    update_curscr()

#endif /* STANDARD_C */

{
	register int  i, j, y, lastpos;

	lastpos = endpos;
	DEC(lastpos);

#ifdef LDEBUG
	LOG(log,"update_curscr   : (%d,%d) to (%d,%d)\n",
		 YX(startpos), YX(lastpos));
	LOG(log,"                : entry color is %s, _color is %s\n",
		 colorname[color], colorname[curscr->_color]);
#endif /* LDEBUG */

	updateflag++;

	if (FORMATTED)
	{
		field = (startpos < firstfield ? screensize - 1 : startpos);
		while (!(screen[field] & FIELDSTART))
			if (--field < 0)
				field = screensize - 1;
		color = colorbits((int)screen[field]);

#ifdef LDEBUG
	LOG(log,"update_curscr   : found field start at (%d,%d), %s\n",
		YX(field), colorname[color]);
#endif /* LDEBUG */

	}
	else
	{
		field = -1;
		color = NORMAL;

#ifdef LDEBUG
	LOG(log,"update_curscr   : unformatted screen \n");
#endif /* LDEBUG */

	}


	i = startpos;
	y = i/linelen;

	/* loop over all of the changed lines, calling makeline */
	do {
		j = (linelen * (y + 1)) - 1;
		if ((lastpos >= i) && (lastpos < j))
			j = lastpos;
		makeline(i,j);
		if (++y >= rows)
			y = 0;
		i = linelen * y;
	} while (j != lastpos);

#ifdef LDEBUG
	LOG(log,"update_curscr   : end with cursor at (%d,%d)\n",
		 ly+1, lx+1);
#endif /* LDEBUG */

}

/*
 * Write the cursor position to the user screen.
 * We assume that the necessary variables are already set.
 */

#ifdef STANDARD_C

extern  void    writecursor(void)

#else /* STANDARD_C */

void    writecursor()

#endif /* STANDARD_C */

{
	domvcur(ly, lx, Y, X);
	ly = Y;
	lx = X;
	f_flush();
	updateflag = 0;
}

/*
 * put the cursor where it belongs.
 * and write it out by a call of writecursor().
 */

#ifdef STANDARD_C

extern  void    putcursor(void)

#else /* STANDARD_C */

void    putcursor()

#endif /* STANDARD_C */

{
	int i;

	X = cursorpos%linelen;
	Y = cursorpos/linelen;
	if (FORMATTED)
		/* set the global variable fieldstart */
		for (i=cursorpos ; ; i = DECR(i))
		{
			if (screen[i] & FIELDSTART)
			{
				fieldstart = i;
				currentcolor = colorbits((int)screen[i]);
				break;
			}
		}
	else
		currentcolor = NORMAL;

#ifdef LDEBUG
	LOG(log, "putcursor       : to (%d,%d), fieldstart to (%d,%d), %s\n",
		      YX(cursorpos), YX(fieldstart), colorname[currentcolor]);
#endif /* LDEBUG */

	writecursor();
}

/* clear the screen into NORMAL mode */

#ifdef STANDARD_C

void    clearnormal(void)

#else /* STANDARD_C */

void    clearnormal()

#endif /* STANDARD_C */

{

#ifdef LDEBUG
	LOG(log, "clearnormal     : current flags is %s\n",
		      colorname[curscr->_color]);
#endif /* LDEBUG */

	f_clearop(-1);

	curscr->_color = NORMAL;
	X = Y = 0;
	lx = ly = 0;
	winerase(curscr);         /* erase the contents of curscr */
}

/* clear the current screen (curscr), but not screen */

#ifdef STANDARD_C

void    clearcurscr(void)

#else /* STANDARD_C */

void    clearcurscr()

#endif /* STANDARD_C */

{
	clearnormal();
	if (slinenum)
	{
		slrefresh();
	}
}

/* reset the data structures to represent a cleared screen */

#ifdef STANDARD_C

void    clearscreen(void)

#else /* STANDARD_C */

void    clearscreen()

#endif /* STANDARD_C */

{
	clearcurscr();
	if (screen)
		(void) memset((void *)screen, 0, (size_t)(screensize * sizeof(short)));
	firstfield = -1;
}

#ifdef SHOWFIELDS

#ifdef STANDARD_C

void    showfields(void)

#else /* STANDARD_C */

void    showfields()

#endif /* STANDARD_C */

{
	/* for debugging */
	int i;
	int fields = 0;

#ifdef LDEBUG
	if (!FORMATTED)
	{
	LOG(log,"showfields      : no fields\n");
		return;
	}
	LOG(log,"showfield       : firstfield at %d,%d ...\n",YX(firstfield));
	for (i=0; i<screensize; i++)
		if (screen[i] & FIELDSTART)
		{
			if (fields++) LOG(log,"->%d,%d\n",YX(i-1));
			LOG(log," Field byte 0x%x (%s%s%s%s%s) %d,%d", screen[i],
			screen[i]&SFprotectedBITS ? " PROT" : "",
			screen[i]&SFnumericBITS ? " NUM" : "",
			screen[i]&SFspecialBITS ? " bLiNk" : "",
			(screen[i]&SFvisibleBITS)==SFinvisibleBITS ? " INVISIBLE"
			    : (screen[i]&SFvisibleBITS)==SFstandoutBITS ? " BRIGHT"
			    : "",
			screen[i]&SFmodifiedBITS ? " MDT" : "",
			YX(i));
		}
	LOG(log,"->%d,%d\n",YX(firstfield-1));
#endif /* LDEBUG */

}
#endif /* SHOWFIELDS */

/* move the cursor "dist" positions (<0 means left).  Make sure
 * the globals "fieldstart" and "currentcolor" are up to date.
 * Note that nothing is yet written: when the cursor is
 * finally positioned then writecursor() is to be called.
 */

#ifdef STANDARD_C

void    movecursor(int dist)

#else /* STANDARD_C */

void    movecursor(dist)

#endif /* STANDARD_C */

{
	register int i;
	register int cp;

	if ((cp = cursorpos + dist) < 0)
		cp += screensize;
	cp = cp % screensize;
	for (i = cp ; i != cursorpos ; i = DECR(i))
	{
		if (screen[i] & FIELDSTART)
		{
			fieldstart = i;
			currentcolor = colorbits((int)screen[i]);
			break;
		}
	}
	X = cp%linelen;
	Y = cp/linelen;
	cursorpos = cp;
}

/*
 *	This routine erases everything on the window.
 *
 * 1/27/81 (Berkeley) @(#)erase.c	1.2
 */

#ifdef STANDARD_C

void    winerase(struct curwin *win)

#else /* STANDARD_C */

void    winerase(win)

struct  curwin  *win;

#endif /* STANDARD_C */

{

	register int            y;

#ifdef DEBUG
	(void)fprintf(outf, "winerase    : WINERASE(%0.2o)\n", win);
#endif /* DEBUG */

	for (y = 0; y < win->_maxy; y++)
	{
		(void) memset((void *)(win->char_y[y]) , ' ' , (size_t)(win->_maxx));
		(void) memset((void *)(win->ccset_y[y]) , 0 , (size_t)(win->_maxx));
	}
	win->_curx = win->_cury = 0;
}
